### PersefoneDesktop - a GUI to the Persefone model of agriculture and ecosystems
###
### This file links the QML UI designs to the underlying model.
###


function loadsimulation(filename)
    global model
    launching[] = true
    fn = convert(filename, String)
    startswith(fn, "file://") && (fn = fn[8:end])
    model = loadmodelobject(fn)
    running[] = false
    landcovermap = rotr90(load(@param(world.landcovermap)))
    date[] = model.date
    progress[] = (date[]-@param(core.startdate)) /
        (@param(core.enddate)-@param(core.startdate))
    launching[] = false
    println("Loaded simulation.")
end

function savesimulation(filename)
    global model
    fn = convert(filename, String)
    startswith(fn, "file://") && (fn = fn[8:end])
    savemodelobject(model, fn)
    println("Saved simulation.")
end

function Base.convert(s::QString, ::Type{<:AbstractString})
    s2 = ""
    for l in s
        s2 *= l
    end
    s2
end

#TODO saveoutput()

function togglerunning()
    global runsimulation, timer
    if running[]
        running[] = false
        QML.stop(timer)
        runbuttontext[] = ">>"
        runbuttontip[] = "Run"
    else
        running[] = true
        runbuttontext[] = "||"
        runbuttontip[] = "Pause"
        QML.start(timer)
    end
end
 
function render_map(screen)
    global model, mapimage, landcovermap
    launching[] && return display(screen, Figure().scene) # blank screen at launch
    println("Updating map")
    #FIXME I'm not sure date-1 is the real fix
    #FIXME repeatedly calling `visualisemap` causes the memory leak
    # (regardless of whether individuals are plotted or just the map)
    #mapimage[] = visualisemap(model, date[]-Day(1), landcovermap)
    empty!(mapimage[])
    mapimage[] = vismap(model, date[]-Day(1), landcovermap)
    display(screen, mapimage[])
end

function vismap(model::AgentBasedModel,date=nothing,landcover=nothing)
    # load and plot the map
    # Note: if the landcover map is supplied, it needs to be rotr90'ed
    isnothing(landcover) && (landcover = load(@param(world.landcovermap)))
    f = Figure()
    ax = Axis(f[1,1])
    hidedecorations!(ax)
    image!(f[1,1], landcover)
    ax.aspect = DataAspect()
    # check if there are individuals and plot them
    inds = model.datatables["individuals"]
    if iszero(size(inds)[1])
        @debug "No individual data to map"
        return f.scene
    end
    isnothing(date) && (date = inds.Date[end])
    for s in unique(inds.Species)
        points = @select!(@subset(inds, :Species .== s, :Date .== date),
                          :X, :Y)
        # The origin in Makie is in the bottom-left rather than in the top-left as
        # on the model map, so we have to invert the Y coordinates
        @transform!(points, :Y = size(model.landscape)[2] .- :Y)
        scatter!(f[1,1], Matrix{Float32}(points), markersize=8)
    end
    f.scene
end

function render_plot(screen)
    #FIXME causes segfault when called during a simulation?
    global model
    launching[] && return display(screen, Figure().scene) # blank screen at launch
    println("Updating plot")
    figure = populationtrends(model)
    display(screen, figure.scene)
end

on(delay) do d
    println("Delay is now $(round(d, digits=1)) seconds.") #XXX DEBUG
end

on(launching) do l
    if l
        @emit showSplash()
    else
        @emit closeSplash()
        @emit updateMakie()
    end
end

datestring = () -> Dates.format(date[], "dd U yyyy")

function activateqmlfunctions()     
    @qmlfunction newsimulation
    @qmlfunction configwindow
    @qmlfunction loadsimulation
    @qmlfunction savesimulation
    @qmlfunction nextstep
    @qmlfunction previousstep
    @qmlfunction togglerunning
    @qmlfunction datestring
    @qmlfunction writeconfig
end

"""
    launch()

The main function that creates the application.
"""
function launch()
    global model, timer

    activateqmlfunctions()
    timer = QTimer()

    GLMakie.activate!()
    ENV["QSG_RENDER_LOOP"] = "basic"
    QML.setGraphicsApi(QML.OpenGL)
    
    qmlfile = joinpath(dirname(@__FILE__), "main.qml")
    loadqml(qmlfile,
            timer = timer,
            vars = JuliaPropertyMap(guiproperties...),
            render_map_callback = @safe_cfunction(render_map, Cvoid, (Any,)),
            render_plot_callback = @safe_cfunction(render_plot, Cvoid, (Any,)))
    launching[] = true
    if isnothing(model)
        #QML.start(timer)
        #newsimulation()
        #QML.stop(timer)
    end
    println("Launched Persefone Desktop.")
    exec() #TODO how do I make this async? -> change to Threads.@spawn?
end
