### PersefoneDesktop - a GUI to the Persefone model of agriculture and ecosystems
###
### This file manages the model initialisation and run functions.
###


function newsimulation()
    global model, landcovermap
    launching[] = true # trigger the launch screen
    running[] = false
    progress[] = 0.0
    model = initialise(configfile[])
    landcovermap = rotr90(load(@param(world.landcovermap)))
    date[] = model.date
    launching[] = false
    println("Model initialised.")
end

function nextstep()
    global model
    if date[] >= @param(core.enddate)
        running[] = false
        runbuttontext[] = ">>"
        runbuttontip[] = "Run"
        return
    elseif date[] < model.date
        # If we've "scrolled back" in time, step forward again without
        # having to resimulate (see `previousstep()`)
        date[] += Day(1)
    else
        # If we're already displaying the newest model update, we need to
        # keep simulating
        stepsimulation!(model)
        date[] = model.date
    end
    progress[] = (date[]-@param(core.startdate)) /
        (@param(core.enddate)-@param(core.startdate))
    @emit updateMakie()
    println("Updated model: $(date[]-Day(1))")
end

function previousstep()
    # Since we store all model output in a dataframe, we can simply
    # "scroll back" in time and visualise the state of the model at
    # a previous date
    global model
    date[] <= @param(core.startdate) && return
    date[] -= Day(1)
    progress[] = (date[]-@param(core.startdate)) /
        (@param(core.enddate)-@param(core.startdate))
    @emit updateMakie()
    println("Backtracked model: $(date[]-Day(1))")
end

#XXX This could also be done with Channels rather than ResumableFunctions,
# which would save us one dependency (Channels are inbuilt). However, the
# ResumableFunctions version seemed to work more smoothly.
@resumable function createrunfunction()
    i = round(delay[]*10)
    while running[]
        i -= 1
        if i <= 0
            nextstep()
            i = round(delay[]*10)
        end
        # to reduce interface lag, always only sleep for 0.1s
        sleep(0.1)
        @yield 1
    end
end


on(running) do r
    ticks = 0 #prevent overflows
    runsimulation = createrunfunction()
    r ? println("Simulation started.") : println("Simulation stopped.") #XXX DEBUG
end

on(ticks) do t
    #launching[] && #FIXME
    running[] && runsimulation()
end
